/*
 * Copyright (C) 2007-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file afs.h Exported symbols of the audio file selector. */

#include <regex.h>
#include "hash.h"

/** Audio file selector data stored in the audio file table. */
struct afs_info {
	/** Seconds since the epoch. */
	uint64_t last_played;
	/** Bit field of set attributes. */
	uint64_t attributes;
	/** Counts how many times the file was selected. */
	uint32_t num_played;
	/** Image blob associated with this file (foreign key). */
	uint32_t image_id;
	/** Lyrics blob associated with this file (foreign key). */
	uint32_t lyrics_id;
	/** Mp3, ogg or aac. */
	uint8_t audio_format_id;
	/** Amplification value. */
	uint8_t amp;
};

/**
 * Events caused by changes to an afs table.
 *
 * Whenever an afs table changes, an event is generated which causes afs to
 * call the event handlers of all other tables. For example, if an audio file
 * is added, the event handler of the mood table checks the new file for
 * admissibility.
 */
enum afs_events {
	/** An attribute was added. */
	ATTRIBUTE_ADD,
	/** An attribute was renamed. */
	ATTRIBUTE_RENAME,
	/** An attribute was removed. */
	ATTRIBUTE_REMOVE,
	/** The afs info struct of an audio file changed. */
	AFSI_CHANGE,
	/** The afh info struct of an audio file changed. */
	AFHI_CHANGE,
	/** An audio file was renamed. */
	AUDIO_FILE_RENAME,
	/** An audio file was added. */
	AUDIO_FILE_ADD,
	/** An audio file is about to be removed. */
	AUDIO_FILE_REMOVE,
	/** A new blob was added. */
	BLOB_ADD,
	/** A blob was renamed. */
	BLOB_RENAME,
	/** A blob is about to be removed. */
	BLOB_REMOVE,
};

/**
 * Used as data for \ref afs_event() for events of type \p ATTRIBUTE_ADD.
 */
struct rmatt_event_data {
	/** The name of the attribute being added. */
	const char *name;
	/** Its bit number. */
	unsigned char bitnum;
};

/**
 * Used as data for \ref afs_event() for events of type \p ATTRIBUTE_AFSI_CHANGE.
 */
struct afsi_change_event_data {
	/** Pointer to the row that has changed. */
	struct osl_row *aft_row;
	/** Afs info before the change. */
	struct afs_info *old_afsi;
};

/** Function pointers for table handling.  */
struct afs_table {
	/** Initializes the other pointers in this struct. */
	void (*init)(struct afs_table *t);
	/** The name of this table. */
	const char *name;
	/** Gets called on startup and on \p SIGHUP. */
	int (*open)(const char *base_dir);
	/** Gets called on shutdown and on \p SIGHUP. */
	void (*close)(void);
	/** Called by the \a init afs command. */
	int (*create)(const char *);
	/** Handles afs events. */
	int (*event_handler)(enum afs_events event, struct para_buffer *pb,
		void *data);
};

/** How audio files are selected by afs. */
enum play_mode {
	/** Admissible files are determined by a mood definition. */
	PLAY_MODE_MOOD,
	/** All listed files are admissible. */
	PLAY_MODE_PLAYLIST,
};

/**
 * Data about one audio file.
 *
 * Needed to produce ls and stat output.
 */
struct ls_data {
	/** Usual audio format handler information. */
	struct afh_info afhi;
	/** Audio file selector information. */
	struct afs_info afsi;
	/** The full path of the audio file. */
	char *path;
	/** The score value (if -a was given). */
	long score;
	/** The sha1 hash of audio file. */
	HASH_TYPE *hash;
};

/** Data about the current audio file, passed from afs to server. */
struct audio_file_data {
	/** The open file descriptor to the current audio file. */
	int fd;
	/** Vss needs this for streaming. */
	struct afh_info afhi;
};

/**
 * Codes used for communication between the server and the afs process.
 *
 * Before forking the afs child, para_server creates a bidirectional pipe
 * through which both processes communicate. Usually para_server requests a new
 * audio in order to start streaming or when the end of the current audio file
 * has been reached.  The afs process responds to such a request by sending
 * back an eight byte buffer. The first four bytes is the uint32_t
 * representation of the code, usually \p NEXT_AUDIO_FILE if an admissible
 * audio file was found, successfully opened and verified. The other four bytes
 * represent the shared memory id of the shared memory area that contains
 * details about the audio file to be streamed next. The open file descriptor
 * of that file is also passed from afs to para_server through the same pipe.
 */
enum afs_server_code {
	/** An audio file was successfully opened. */
	NEXT_AUDIO_FILE,
	/** No admissible audio file was found. */
	NO_ADMISSIBLE_FILES,
};

/** Flags passed to for_each_matching_row(). */
enum pattern_match_flags {
	/** Loop in reverse order. */
	PM_REVERSE_LOOP = 1,
	/** If no pattern is given, loop over all rows. */
	PM_NO_PATTERN_MATCHES_EVERYTHING = 2,
	/** If the data in match_column is the empty string, skip this row. */
	PM_SKIP_EMPTY_NAME = 4,
};

/** Structure passed to for_each_matching_row(). */
struct pattern_match_data {
	/** Loop over all rows in this table. */
	struct osl_table *table;
	/** Determines the loop order. Must be an rbtree column. */
	unsigned loop_col_num;
	/** Data from this column is matched against the given patterns. */
	unsigned match_col_num;
	/** \see pattern_match_flags. */
	unsigned pm_flags;
	/** This value is passed verbatim to fnmatch(). */
	int fnmatch_flags;
	/** Null-terminated array of patterns. */
	struct osl_object patterns;
	/** Data pointer passed to the action function. */
	void *data;
	/** For each matching row, this function will be called. */
	int (*action)(struct osl_table *table, struct osl_row *row, const char *name, void *data);
};


/**
 * Afs command handlers run as a process which is not related to the afs
 * process, i.e. they can not change the address space of afs directly.
 * Therefore afs commands typically consist of two functions: The command
 * handler and the corresponding callback function that runs in afs context.
 *
 * \sa send_callback_request().
 */
typedef void callback_function(int fd, const struct osl_object *);

/**
 * Callbacks send chunks to data back to the command handler. Pointers to
 * this type of function are used by \ref send_callback_request and friends
 * to deal with the data in the command handler process.
 *
 * \sa \ref send_callback_request().
 */
typedef int callback_result_handler(struct osl_object *result, void *private);
int rc4_send_result(struct osl_object *result, void *private);
int pass_buffer_as_shm(char *buf, size_t size, void *fd_ptr);

__noreturn void afs_init(uint32_t cookie, int socket_fd);
void afs_event(enum afs_events event, struct para_buffer *pb,
	void *data);
int send_callback_request(callback_function *f, struct osl_object *query,
		callback_result_handler *result_handler,
		void *private_result_data);
int send_option_arg_callback_request(struct osl_object *options,
		int argc,  char * const * const argv, callback_function *f,
		callback_result_handler *result_handler,
		void *private_result_data);
int send_standard_callback_request(int argc,  char * const * const argv,
		callback_function *f, callback_result_handler *result_handler,
		void *private_result_data);
int string_compare(const struct osl_object *obj1, const struct osl_object *obj2);
int for_each_matching_row(struct pattern_match_data *pmd);

/* score */
void score_init(struct afs_table *t);
int admissible_file_loop(void *data, osl_rbtree_loop_func *func);
int admissible_file_loop_reverse(void *data, osl_rbtree_loop_func *func);
int score_get_best(struct osl_row **aft_row, long *score);
int get_score_and_aft_row(struct osl_row *score_row, long *score, struct osl_row **aft_row);
int score_add(const struct osl_row *row, long score);
int score_update(const struct osl_row *aft_row, long new_score);
int get_num_admissible_files(unsigned *num);
int score_delete(const struct osl_row *aft_row);
int clear_score_table(void);
int row_belongs_to_score_table(const struct osl_row *aft_row, unsigned *rank);

/* attribute */
void attribute_init(struct afs_table *t);
void get_attribute_bitmap(const uint64_t *atts, char *buf); /* needed by com_ls() */
int get_attribute_bitnum_by_name(const char *att_name, unsigned char *bitnum);
int get_attribute_text(uint64_t *atts, const char *delim, char **text);

/* aft */
void aft_init(struct afs_table *t);
int aft_get_row_of_path(const char *path, struct osl_row **row);
int open_and_update_audio_file(struct osl_row *aft_row, long score,
	struct audio_file_data *afd);
int load_afd(int shmid, struct audio_file_data *afd);
int load_afsi(struct afs_info *afsi, struct osl_object *obj);
void save_afsi(struct afs_info *afsi, struct osl_object *obj);
int get_afsi_of_row(const struct osl_row *row, struct afs_info *afsi);
int get_afhi_of_row(const struct osl_row *row, struct afh_info *afhi);
int get_afsi_of_path(const char *path, struct afs_info *afsi);
int get_audio_file_path_of_row(const struct osl_row *row, char **path);
int get_afsi_object_of_row(const struct osl_row *row, struct osl_object *obj);
int audio_file_loop(void *private_data, osl_rbtree_loop_func *func);
void aft_check_callback(int fd, __a_unused const struct osl_object *query);

/* playlist */
int playlist_open(char *name);
void playlist_close(void);
void playlist_check_callback(int fd, __a_unused const struct osl_object *query);

/** evaluates to 1 if x < y, to -1 if x > y and to 0 if x == y */
#define NUM_COMPARE(x, y) ((int)((x) < (y)) - (int)((x) > (y)))


/** Define exported functions and a table pointer for an osl blob table. */
#define DECLARE_BLOB_SYMBOLS(table_name, cmd_prefix) \
	void table_name ## _init(struct afs_table *t); \
	int cmd_prefix ## _get_name_by_id(uint32_t id, char **name); \
	int cmd_prefix ## _get_def_by_id(uint32_t id, struct osl_object *def); \
	int cmd_prefix ## _get_def_by_name(char *name, struct osl_object *def); \
	int cmd_prefix ## _get_name_and_def_by_row(const struct osl_row *row, \
		char **name, struct osl_object *def); \
	int table_name ##_event_handler(enum afs_events event, \
		struct para_buffer *pb, void *data); \
	extern struct osl_table *table_name ## _table;

DECLARE_BLOB_SYMBOLS(lyrics, lyr);
DECLARE_BLOB_SYMBOLS(images, img);
DECLARE_BLOB_SYMBOLS(moods, mood);
DECLARE_BLOB_SYMBOLS(playlists, pl);

/** The columns of an abstract blob table. */
enum blob_table_columns {
	/** The identifier, a positive integer that never repeats. */
	BLOBCOL_ID,
	/** The unique name of the blob. */
	BLOBCOL_NAME,
	/** The actual blob contents. */
	BLOBCOL_DEF,
	/** A blob table has that many columns. */
	NUM_BLOB_COLUMNS
};
