/*
 * Copyright (C) 1997-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file client_common.c Common functions of para_client and para_audiod. */

#include <regex.h>
#include <sys/types.h>
#include <dirent.h>
#include <openssl/rc4.h>

#include "para.h"
#include "error.h"
#include "list.h"
#include "sched.h"
#include "client.cmdline.h"
#include "crypt.h"
#include "rc4.h"
#include "net.h"
#include "fd.h"
#include "string.h"
#include "client.cmdline.h"
#include "client.h"
#include "hash.h"

/**
 * Close the connection to para_server and free all resources.
 *
 * \param ct Pointer to the client data.
 *
 * \sa client_open.
 */
void client_close(struct client_task *ct)
{
	if (!ct)
		return;
	if (ct->rc4c.fd >= 0)
		close(ct->rc4c.fd);
	free(ct->buf);
	free(ct->user);
	free(ct->config_file);
	free(ct->key_file);
	client_cmdline_parser_free(&ct->conf);
	free(ct);
}

/**
 * The preselect hook for server commands.
 *
 * \param s Pointer to the scheduler.
 * \param t Pointer to the task struct for this command.
 *
 * The task pointer must contain a pointer to the initialized client data
 * structure as it is returned by client_open().
 *
 * This function checks the state of the connection and adds the file descriptor
 * of the connection to the read or write fd set of \a s accordingly.
 *
 * \sa register_task() client_open(), struct sched, struct task.
 */
static void client_pre_select(struct sched *s, struct task *t)
{
	struct client_task *ct = container_of(t, struct client_task, task);

	ct->check_r = 0;
	ct->check_w = 0;
	if (ct->rc4c.fd < 0)
		return;
	switch (ct->status) {
	case CL_CONNECTED:
	case CL_SENT_AUTH:
	case CL_SENT_CH_RESPONSE:
	case CL_SENT_COMMAND:
		para_fd_set(ct->rc4c.fd, &s->rfds, &s->max_fileno);
		ct->check_r = 1;
		return;

	case CL_RECEIVED_WELCOME:
	case CL_RECEIVED_CHALLENGE:
	case CL_RECEIVED_PROCEED:
		para_fd_set(ct->rc4c.fd, &s->wfds, &s->max_fileno);
		ct->check_w = 1;
		return;

	case CL_RECEIVING:
		if (ct->loaded < CLIENT_BUFSIZE - 1) {
			para_fd_set(ct->rc4c.fd, &s->rfds, &s->max_fileno);
			ct->check_r = 1;
		}
		return;
	case CL_SENDING:
		if (!ct->in_loaded) /* stdin task not yet started */
			return;
		if (*ct->in_loaded) {
			PARA_INFO_LOG("loaded: %zd\n", *ct->in_loaded);
			para_fd_set(ct->rc4c.fd, &s->wfds, &s->max_fileno);
			ct->check_w = 1;
		} else {
			if (*ct->in_error) {
				t->error = *ct->in_error;
				s->timeout.tv_sec = 0;
				s->timeout.tv_usec = 1;
			}
		}
		return;
	}
}

static ssize_t client_recv_buffer(struct client_task *ct)
{
	ssize_t ret;

	if (ct->status < CL_SENT_CH_RESPONSE)
		ret = recv_buffer(ct->rc4c.fd, ct->buf + ct->loaded,
			CLIENT_BUFSIZE - ct->loaded);
	else
		ret = rc4_recv_buffer(&ct->rc4c, ct->buf + ct->loaded,
			CLIENT_BUFSIZE - ct->loaded);
	if (!ret)
		return -E_SERVER_EOF;
	if (ret > 0)
		ct->loaded += ret;
	return ret;
}

/**
 * The post select hook for client commands.
 *
 * \param s Pointer to the scheduler.
 * \param t Pointer to the task struct for this command.
 *
 * Depending on the current state of the connection and the status of the read
 * and write fd sets of \a s, this function performs the necessary steps to
 * authenticate the connection, to send the command given by \a t->private_data
 * and to receive para_server's output, if any.
 *
 * \sa struct sched, struct task.
 */
static void client_post_select(struct sched *s, struct task *t)
{
	struct client_task *ct = container_of(t, struct client_task, task);

	t->error = 0;
	if (ct->rc4c.fd < 0)
		return;
	if (!ct->check_r && !ct->check_w)
		return;
	if (ct->check_r && !FD_ISSET(ct->rc4c.fd, &s->rfds))
		return;
	if (ct->check_w && !FD_ISSET(ct->rc4c.fd, &s->wfds))
		return;
	switch (ct->status) {
	case CL_CONNECTED: /* receive welcome message */
		t->error = client_recv_buffer(ct);
		if (t->error < 0)
			goto err;
		ct->status = CL_RECEIVED_WELCOME;
		return;
	case CL_RECEIVED_WELCOME: /* send auth command */
		sprintf(ct->buf, AUTH_REQUEST_MSG "%s", ct->user);
		PARA_INFO_LOG("--> %s\n", ct->buf);
		t->error = send_buffer(ct->rc4c.fd, ct->buf);
		if (t->error < 0)
			goto err;
		ct->status = CL_SENT_AUTH;
		return;
	case CL_SENT_AUTH: /* receive challenge and rc4 keys */
		ct->loaded = 0;
		t->error = client_recv_buffer(ct);
		if (t->error < 0)
			goto err;
		ct->loaded = t->error;
		PARA_INFO_LOG("<-- [challenge] (%zu bytes)\n", ct->loaded);
		ct->status = CL_RECEIVED_CHALLENGE;
		return;
	case CL_RECEIVED_CHALLENGE:
		{
		/* decrypted challenge/rc4 buffer */
		unsigned char crypt_buf[1024];
		/* the SHA1 of the decrypted challenge */
		unsigned char challenge_sha1[HASH_SIZE];

		t->error = para_decrypt_buffer(ct->key_file, crypt_buf,
			(unsigned char *)ct->buf, ct->loaded);
		if (t->error < 0)
			goto err;
		sha1_hash((char *)crypt_buf, CHALLENGE_SIZE, challenge_sha1);
		RC4_set_key(&ct->rc4c.send_key, RC4_KEY_LEN,
			crypt_buf + CHALLENGE_SIZE);
		RC4_set_key(&ct->rc4c.recv_key, RC4_KEY_LEN,
			crypt_buf + CHALLENGE_SIZE + RC4_KEY_LEN);
		hash_to_asc(challenge_sha1, ct->buf);
		PARA_INFO_LOG("--> %s\n", ct->buf);
		t->error = send_bin_buffer(ct->rc4c.fd, (char *)challenge_sha1,
			HASH_SIZE);
		if (t->error < 0)
			goto err;
		ct->status = CL_SENT_CH_RESPONSE;
		return;
		}
	case CL_SENT_CH_RESPONSE: /* read server response */
		{
		size_t bytes_received;
		ct->loaded = 0;
		t->error = client_recv_buffer(ct);
		if (t->error < 0)
			goto err;
		bytes_received = t->error;
		/* check if server has sent "Proceed" message */
		t->error = -E_CLIENT_AUTH;
		if (bytes_received < PROCEED_MSG_LEN)
			goto err;
		if (!strstr(ct->buf, PROCEED_MSG))
			goto err;
		ct->status = CL_RECEIVED_PROCEED;
		t->error = 0;
		return;
		}
	case CL_RECEIVED_PROCEED: /* concat args and send command */
		{
		int i;
		char *command = NULL;
		for (i = 0; i < ct->conf.inputs_num; i++) {
			char *tmp = command;
			command = make_message("%s\n%s", command?
				command : "", ct->conf.inputs[i]);
			free(tmp);
		}
		command = para_strcat(command, EOC_MSG "\n");
		PARA_DEBUG_LOG("--> %s\n", command);
		t->error = rc4_send_buffer(&ct->rc4c, command);
		free(command);
		if (t->error < 0)
			goto err;
		ct->status = CL_SENT_COMMAND;
		return;
		}
	case CL_SENT_COMMAND:
		ct->loaded = 0;
		t->error = client_recv_buffer(ct);
		if (t->error < 0)
			goto err;
		if (strstr(ct->buf, AWAITING_DATA_MSG))
			ct->status = CL_SENDING;
		else
			ct->status = CL_RECEIVING;
		return;
	case CL_SENDING:
		PARA_INFO_LOG("loaded: %zd\n", *ct->in_loaded);
		t->error = rc4_send_bin_buffer(&ct->rc4c, ct->inbuf,
			*ct->in_loaded);
		if (t->error < 0)
			goto err;
		*ct->in_loaded = 0;
		return;
	case CL_RECEIVING:
		t->error = client_recv_buffer(ct);
		if (t->error < 0)
			goto err;
		return;
	}
err:
	if (t->error != -E_SERVER_EOF)
		PARA_ERROR_LOG("%s\n", para_strerror(-t->error));
}

/* connect to para_server and register the client task */
static int client_connect(struct client_task *ct)
{
	int ret;

	ct->rc4c.fd = -1;
	ret = makesock(AF_UNSPEC, IPPROTO_TCP, 0, ct->conf.hostname_arg,
		ct->conf.server_port_arg);
	if (ret < 0)
		return ret;
	ct->rc4c.fd = ret;
	ct->status = CL_CONNECTED;
	ret = mark_fd_nonblocking(ct->rc4c.fd);
	if (ret < 0)
		goto err_out;
	ct->task.pre_select = client_pre_select;
	ct->task.post_select = client_post_select;
	sprintf(ct->task.status, "client");
	register_task(&ct->task);
	return 1;
err_out:
	close(ct->rc4c.fd);
	ct->rc4c.fd = -1;
	return ret;
}

/**
 * Open connection to para_server.
 *
 * \param argc Usual argument count.
 * \param argv Usual argument vector.
 * \param ct_ptr Points to dynamically allocated and initialized client task
 * struct upon successful return.
 * \param loglevel If not \p NULL, the number of the loglevel is stored here.
 *
 * Check the command line options given by \a argc and argv, set default values
 * for user name and rsa key file, read further option from the config file.
 * Finally, establish a connection to para_server.
 *
 * \return Standard.
 */
int client_open(int argc, char *argv[], struct client_task **ct_ptr,
		int *loglevel)
{
	char *home = para_homedir();
	int ret;
	struct client_task *ct = para_calloc(sizeof(struct client_task));

	ct->buf = para_malloc(CLIENT_BUFSIZE);
	*ct_ptr = ct;
	ct->rc4c.fd = -1;
	ret = -E_CLIENT_SYNTAX;
	if (client_cmdline_parser(argc, argv, &ct->conf))
		goto out;
	HANDLE_VERSION_FLAG("client", ct->conf);
	ret = -E_CLIENT_SYNTAX;
	if (!ct->conf.inputs_num)
		goto out;
	ct->user = ct->conf.user_given?
		para_strdup(ct->conf.user_arg) : para_logname();

	ct->key_file = ct->conf.key_file_given?
		para_strdup(ct->conf.key_file_arg) :
		make_message("%s/.paraslash/key.%s", home, ct->user);

	ct->config_file = ct->conf.config_file_given?
		para_strdup(ct->conf.config_file_arg) :
		make_message("%s/.paraslash/client.conf", home);
	ret = file_exists(ct->config_file);
	if (!ret && ct->conf.config_file_given) {
		ret = -E_NO_CONFIG;
		goto out;
	}
	if (ret) {
		struct client_cmdline_parser_params params = {
			.override = 0,
			.initialize = 0,
			.check_required = 0,
			.check_ambiguity = 0,
			.print_errors = 0
		};
		ret = -E_BAD_CONFIG;
		if (client_cmdline_parser_config_file(ct->config_file,
			&ct->conf, &params))
			goto out;
	}
	if (loglevel)
		*loglevel = get_loglevel_by_name(ct->conf.loglevel_arg);
	PARA_INFO_LOG("loglevel: %s\n", ct->conf.loglevel_arg);
	PARA_INFO_LOG("config_file: %s\n", ct->config_file);
	PARA_INFO_LOG("key_file: %s\n", ct->key_file);
	PARA_NOTICE_LOG("connecting %s:%d\n", ct->conf.hostname_arg,
		ct->conf.server_port_arg);
	ret = client_connect(ct);
out:
	free(home);
	if (ret < 0) {
		PARA_ERROR_LOG("%s\n", para_strerror(-ret));
		client_close(ct);
		*ct_ptr = NULL;
	}
	return ret;
}
