/*
 * Copyright (C) 2006-2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file write_common.c common functions of para_audiod and para_write */

#include <regex.h>

#include "para.h"
#include "string.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "write.h"
#include "error.h"

/** the array containing the names of all supported writers */
const char *writer_names[] ={WRITER_NAMES};

/** the array of supported writers */
struct writer writers[NUM_SUPPORTED_WRITERS] = {WRITER_ARRAY};

static void wng_pre_select(struct sched *s, struct task *t)
{
	struct writer_node_group *g = container_of(t, struct writer_node_group, task);
	int i;

	FOR_EACH_WRITER_NODE(i, g) {
		struct writer_node *wn = &g->writer_nodes[i];
		struct writer *w = writers + wn->writer_num;
		if (!w->pre_select)
			continue;
		t->error = w->pre_select(s, wn);
		if (t->error < 0)
			return;
	}
	/*
	 * Force a minimal delay if something was written during the previous
	 * call to wng_post_select(). This is necessary because the filter
	 * chain might still have data for us which it couldn't convert during
	 * the previous run due to its buffer size constraints. In this case we
	 * do not want to wait until the next input data arrives as this could
	 * lead to buffer underruns.
	 */
	if (g->last_written == 0)
		return;
	s->timeout.tv_sec = 0;
	s->timeout.tv_usec = 1;
}

static void wng_post_select(struct sched *s, struct task *t)
{
	struct writer_node_group *g = container_of(t, struct writer_node_group, task);
	int i;
	size_t min_written = 0, max_written = 0;

	FOR_EACH_WRITER_NODE(i, g) {
		struct writer_node *wn = &g->writer_nodes[i];
		struct writer *w = writers + wn->writer_num;
		t->error = w->post_select(s, wn);
		if (t->error < 0)
			return;
		if (!i)
			min_written = wn->written;
		else
			min_written = PARA_MIN(min_written, wn->written);
		max_written = PARA_MAX(max_written, wn->written);
	}
	g->last_written = max_written;
	//PARA_INFO_LOG("loaded: %zd, min_written: %zd bytes\n", *g->loaded, min_written);
	if (min_written) {
		*g->loaded -= min_written;
		FOR_EACH_WRITER_NODE(i, g)
			g->writer_nodes[i].written -= min_written;
	}
	if (!*g->loaded && *g->input_error) {
		t->error = *g->input_error;
		return;
	}
	if (*g->loaded && min_written) {
//		PARA_INFO_LOG("moving %zd bytes\n", *g->loaded);
		memmove(*g->bufp, *g->bufp + min_written, *g->loaded);
	}
}

/**
 * call the open function of each writer in the group
 *
 * \param g the writer node group
 *
 * \return If at least one open function returned an error, all successful
 * writer notes get closed and this error value is returned. Upon success, a
 * task associated with \a g is registered to the scheduler and the function
 * returns a positive value.
 * */
int wng_open(struct writer_node_group *g)
{
	int i, ret = 1;

	PARA_NOTICE_LOG("opening wng %p with %d writer(s)\n", g, g->num_writers);
	FOR_EACH_WRITER_NODE(i, g) {
		struct writer_node *wn = &g->writer_nodes[i];
		struct writer *w = writers + wn->writer_num;
		wn->wng = g;
		ret = w->open(wn);
		if (ret < 0)
			goto err_out;
	}
	sprintf(g->task.status, "%s", "writer node group");
	register_task(&g->task);
	g->open = 1;
	return 1;
err_out:
	PARA_ERROR_LOG("%s\n", para_strerror(-ret));
	while (i > 0) {
		struct writer_node *wn = &g->writer_nodes[--i];
		struct writer *w = writers + wn->writer_num;
		w->close(wn);
	}
	free(g->writer_nodes);
	g->num_writers = 0;
	g->task.error = -E_TASK_UNREGISTERED;
	return ret;
}

/**
 * call the close function of each writer in the given group
 *
 * \param g the writer node group to close
 *
 * This function also frees all resources of the given group.
 */
void wng_close(struct writer_node_group *g)
{
	int i;

	if (!g || !g->open)
		return;
	PARA_NOTICE_LOG("closing wng with %d writer(s)\n", g->num_writers);
	FOR_EACH_WRITER_NODE(i, g) {
		struct writer_node *wn = &g->writer_nodes[i];
		struct writer *w = writers + wn->writer_num;
		w->close(wn);
	}
	free(g->writer_nodes);
	free(g);
}

/**
 * allocate and initialize a new writer_node_group struct
 *
 * \param num_writers the number of writer nodes for the new group
 *
 * \return Pointer to the new writer node group
 */
struct writer_node_group *wng_new(unsigned num_writers)
{
	struct writer_node_group *g = para_calloc(sizeof(struct writer_node_group));
	g->num_writers = num_writers;
	g->writer_nodes = para_calloc(num_writers
		* sizeof(struct writer_node));
	g->task.post_select = wng_post_select;
	g->task.pre_select = wng_pre_select;
	return g;
}

/**
 * Call the init function of each supported paraslash writer.
 */
void writer_init(void)
{
	int i;

	FOR_EACH_WRITER(i)
		writers[i].init(&writers[i]);
}
/**
 * check if given string is a valid command line for any writer
 *
 * \param \wa string of the form writer_name:options
 * \param writer_num contains the number of the writer upon success
 *
 * This function checks whether \a wa starts with the name of a supported
 * paraslash writer, optionally followed by a colon and any options for that
 * writer.  If a valid writer name was found and further are present, the
 * remaining part of \a wa is passed to that writer's config parser.
 *
 * \return On success, a pointer to the gengetopt args info struct is returned
 * and \a writer_num contains the number of the writer. Otherwise this function
 * returns \p NULL.
 */
void *check_writer_arg(const char *wa, int *writer_num)
{
	int i;

	*writer_num = -E_WRITE_COMMON_SYNTAX;
	PARA_INFO_LOG("checking  %s\n", wa);
	FOR_EACH_WRITER(i) {
		const char *name = writer_names[i];
		size_t len = strlen(name);
		char c;
		if (strlen(wa) < len)
			continue;
		if (strncmp(name, wa, len))
			continue;
		c = wa[len];
		if (c && c != ' ')
			continue;
		if (c && !writers[i].parse_config)
			return NULL;
		*writer_num = i;
		return writers[i].parse_config(c? wa + len + 1 : "");
	}
	PARA_ERROR_LOG("writer not found\n");
	return NULL;
}

/**
 * setup a writer node group with only one writer, the default writer
 *
 * The writer which is set up depends on the OS. It defaults to alsa for Linux,
 * osx_write for OS X, file writer if neither of these is supported.
 *
 * \return pointer to the allocated writer node group
 */
struct writer_node_group *setup_default_wng(void)
{
	struct writer_node_group *wng = wng_new(1);
	wng->writer_nodes[0].writer_num = DEFAULT_WRITER;
	PARA_INFO_LOG("using default writer: %s %p\n",
		writer_names[DEFAULT_WRITER], writers[DEFAULT_WRITER].parse_config);
	wng->writer_nodes[0].conf = writers[DEFAULT_WRITER].parse_config("");
	return wng;
}
/**
 * Print the help text of all writers to stdout.
 *
 * \param detailed Whether to print the detailed help text.
 */
void print_writer_helps(int detailed)
{
	int i;

	printf_or_die("\nAvailable writers: \n\t");
	FOR_EACH_WRITER(i)
		printf_or_die("%s%s", i? " " : "", writer_names[i]);
	printf_or_die("\n\n");
	FOR_EACH_WRITER(i) {
		struct writer *w = writers + i;

		if (!w->help.short_help)
			continue;
		printf_or_die("Options for %s:\n", writer_names[i]);
		ggo_print_help(&w->help, detailed);
	}
}
